#include "ext.h"
#include "z_dsp.h"
#include <stdlib.h>
// #include "stdio.h"

void *granola_class;



typedef struct _granola
{
    t_pxobject x_obj;
	float *gbuf;
  	int grainsamps;
  	int buflen ;
  	float *grainenv; 

	int gpt1;
	int gpt2;
	int gpt3;
	float phs1;
	float phs2;
	float phs3;
	float incr;
	int curdel;
	short mute_me;
	short iconnect;

} t_granola;

void *granola_new(double val);
t_int *offset_perform(t_int *w);
t_int *granola_perform(t_int *w);
void granola_dsp(t_granola *x, t_signal **sp, short *count);
void granola_assist(t_granola *x, void *b, long m, long a, char *s);
void granola_mute(t_granola *x, t_symbol *msg, short argc, t_atom *argv);
void granola_float(t_granola *x, double f ) ;

void main(void)
{
    setup((t_messlist **)&granola_class, (method)granola_new, (method)dsp_free, (short)sizeof(t_granola), 0L, A_DEFFLOAT, 0);
    addmess((method)granola_dsp, "dsp", A_CANT, 0);
    addmess((method)granola_assist,"assist",A_CANT,0);
    addmess((method)granola_mute,"mute",A_GIMME,0);
	addfloat((method)granola_float);
    dsp_initclass();
	post("granola~ by Eric Lyon");
}

void granola_float(t_granola *x, double f) {
	x->incr = f; 
} 

void granola_mute(t_granola *x, t_symbol *msg, short argc, t_atom *argv)
{
	x->mute_me = argv[0].a_w.w_long;
}
void granola_assist (t_granola *x, void *b, long msg, long arg, char *dst)
{
	if (msg==1) {
		switch (arg) {
			case 0:sprintf(dst,"(signal) Input ");break;
			case 1:sprintf(dst,"(signal/float) Increment ");break;
		}
	} else if (msg==2) {
		sprintf(dst,"(signal) Output ");
	}
}

void *granola_new(double val)
{
    t_granola *x = (t_granola *)newobject(granola_class);

  	int i;


    dsp_setup((t_pxobject *)x,2);
    outlet_new((t_pxobject *)x, "signal");
// INITIALIZATIONS
	if( val > 0 ) {
		x->grainsamps = val ;
		// post( "grainsize set to %.0f", val );
	} else {
		x->grainsamps = 2048 ;
		// post( "grainsize defaults to %d, val was %.0f", x->grainsamps, val );

	}
	x->buflen = x->grainsamps * 4;
  	x->gbuf = (float *) calloc( x->buflen, sizeof(float) ) ;
  	x->grainenv = (float *) calloc( x->grainsamps, sizeof(float) );
  	for(i = 0; i < x->grainsamps; i++ ){
   	 x->grainenv[i] = .5 + (-.5 * cos( TWOPI * ((float)i/(float)x->grainsamps) ) );
  	}
  	x->gpt1 = 0;
  	x->gpt2 = x->grainsamps / 3.;
  	x->gpt3 = 2. * x->grainsamps / 3.;
  	x->phs1 = 0;
  	x->phs2 = x->grainsamps / 3. ;
  	x->phs3 = 2. * x->grainsamps / 3. ;
  	x->incr = .5 ;
  	x->curdel = 0;
  	x->mute_me = 0;

    return (x);
}

t_int *granola_perform(t_int *w)
{
	float sample, outsamp ;
	int iphs_a, iphs_b;
	float m1, m2;
  	double floor(double x);
	
	/****/
	t_granola *x = (t_granola *) (w[1]);
	t_float *in = (t_float *)(w[2]);
	t_float *increment = (t_float *)(w[3]);
	t_float *out = (t_float *)(w[4]);
	int n = (int)(w[5]);
	int iconnect = x->iconnect;
if( x->mute_me ) {
	while( n-- ){
		*out++ = 0.0;
	}
	return (w+6);
} else {	
	
	while (n--) { 
		if( iconnect )
			x->incr = *increment++ ;
		if( x->incr <= 0. ) {
			x->incr = .5 ;
		}
		
  	 	if( x->curdel >= x->buflen ){
      		x->curdel = 0 ;
    	}    
    	x->gbuf[ x->curdel ] = *in++;
    	
	// grain 1 
    iphs_a = floor( x->phs1 );
    iphs_b = iphs_a + 1;

    m2 = x->phs1 - iphs_a;
    m1 = 1. - m2 ;
    while( iphs_a >= x->buflen ) {
      iphs_a -= x->buflen;
    }
    while( iphs_b >= x->buflen ) {
      iphs_b -= x->buflen;
    }
    outsamp = ( m1 * x->gbuf[ iphs_a ] + m2 * x->gbuf[ iphs_b ] ) * x->grainenv[ x->gpt1 ];
    ++(x->gpt1);
    if( x->gpt1 >= x->grainsamps ) {

      x->gpt1 = 0;
      x->phs1 = x->curdel ;
    }
    x->phs1 += x->incr;
    while( x->phs1 >= x->buflen ) {
      x->phs1 -= x->buflen ;
    }
    
   // now add second grain 


    iphs_a = floor( x->phs2 );
    
    iphs_b = iphs_a + 1;
    
    m2 = x->phs2 - iphs_a;
    
    m1 = 1. - m2 ;
    
    while( iphs_a >= x->buflen ) {
      iphs_a -= x->buflen;
    }
    while( iphs_b >= x->buflen ) {
      iphs_b -= x->buflen;
    }
				
    outsamp += ( m1 * x->gbuf[ iphs_a ] + m2 * x->gbuf[ iphs_b ] ) * x->grainenv[ x->gpt2 ];

    ++(x->gpt2);
    if( x->gpt2 >= x->grainsamps ) {

      x->gpt2 = 0;
      x->phs2 = x->curdel ;
    }
    x->phs2 += x->incr ;    
    while( x->phs2 >= x->buflen ) {
      x->phs2 -= x->buflen ;
    }

    // now add third grain 

    iphs_a = floor( x->phs3 );
    iphs_b = iphs_a + 1;
    m2 = x->phs3 - iphs_a;
    m1 = 1. - m2 ;
    while( iphs_a >= x->buflen ) {
      iphs_a -= x->buflen;
    }
    while( iphs_b >= x->buflen ) {
      iphs_b -= x->buflen;
    }
				
    outsamp += ( m1 * x->gbuf[ iphs_a ] + m2 * x->gbuf[ iphs_b ] ) * x->grainenv[ x->gpt3 ];

    ++(x->gpt3);
    if( x->gpt3 >= x->grainsamps ) {
      x->gpt3 = 0;
      x->phs3 = x->curdel ;
    }
    x->phs3 += x->incr ;    
    while( x->phs3 >= x->buflen ) {
      x->phs3 -= x->buflen ;
    }


    ++(x->curdel);
    
            
	*out++ = outsamp; 
	/* output may well need to attenuated */
	}
	return (w+6);
	}
}		

void granola_dsp(t_granola *x, t_signal **sp, short *count)
{
	long i;
	x->iconnect = count[1];

		dsp_add(granola_perform, 5, x, sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,  sp[0]->s_n);
}

